/**
 * Experimental module, not for distribution. 
 * Rush-Larsen solver for cardiac excitability kinetic models. 
 * Utilizes a special kinetics format, describing HH-type gates separately. 
 * This version uses the order of calculations: tabulated functions - other - gates.
 */

#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "system.h"
#include "beatbox.h"
#include "device.h"
#include "state.h"
#include "bikt.h"
#include "ionic.h"
#include "qpp.h"

typedef struct {
  real ht;			/* time step */
  Name order;			/* text code of the order of execution */
  int whichorder;		/* numeric code of the order of execution */
  Name ionic;			/* ionic of the ionic cell model */
  ionic_str I;			/* ionic cell description */
  int exp_ngate;		/* if nonzero, non-tabulated gates are steppedn by RL, otherwise FE */
  int rest;			/* how many steps to do to find resting values */
  real Vmin, Vmax;		/* limits of voltage in the table (Vmax is approximate) */
  real dV;			/* voltage increment in the table */
  int equilibrate_gates;	/* whether to equilibrate gates in the initial state */
  real *u;			/* vector of vars for steady state if any */
  real *du;			/* vector of derivatives */
  real *nalp;			/* vector of nontab alphas */
  real *nbet;			/* vector of nontab betas */
  int nV;			/* number of rows in the table */
  real one_o_dV;		/* inverse of voltage increment in the table */
  real *tab;			/* the table of values of functions */
  real *adhoc;			/* array of ad hoc values of tabulable functions */
} STR;

/************************************************/
/* Structure of the state vector:		*/
/* 0 .. no-1: no "other variables"		*/
/* no .. no+nn-1: nn non-tab gates      	*/
/* no+nn .. no+nn+nt-1=nv-1":  ng gates		*/
/* So nv=no+nn+nt				*/

/****************************************************************/
/* Structure of the vector of tabulated functions: 		*/
/* 0 .. nt-1 		alphas/a-coefficients	   		*/
/* nt .. 2*nt-1 	betas/b-coefficients	   		*/
/* 2*nt .. ntab-1	everything else		   		*/
/* NB: the first 2*nt values are dual purpose (alp/bet->a/b)	*/

/********************************************************/
/* Conversion of alpha anb beta into the		*/
/* Rush-Larsen linear combination coefficients: 	*/
/* a=alpha/(alpha+beta)*(1-exp(-(alpha+beta)*ht)); 	*/
/* b=exp(-(alpha+beta)*ht). 			        */
/* This is a bit tricky; for debugging might be an idea */
/* to do this in a more straightforward way...          */
static inline void calcab (real *a,real *b,real ht) { 
  /* first a=alp, b=bet */
  (*b)+=(*a);	   /* b=alp+bet         */
  (*a)/=(*b);	   /* a=alp/(alp+bet) */
  (*b)*=ht;	   /* b=(alp+bet)*ht) */
  (*b)=exp(-(*b)); /* b=exp(-(alp+bet)*ht) */
  (*a)*=(1-(*b));  /* a=alp/(alp+bet)*(1-exp(-(alp+bet)*ht)) */
}

/* Substeps of the algorithm can be done in different order */

/***********************/
/* TABULATED FUNCTIONS */
/* check voltage */							
#define DOTABLES \
  V=u[V_index];								\
  iV=floor((V-Vmin)*one_o_dV);						\
  /* if outside limits calculate tabulated values by formulas */	\
  if (iV<0 || iV>=nV) {							\
    values=adhoc;							\
    if (!ftab(V,values,ntab))						\
      ABORT("error calculating ftab(%s) at t=%ld point %d %d %d: V=%g\n",ionic,t,x,y,z,V); \
    for (it=0;it<nt;it++)						\
      calcab(values+it,values+nt+it,ht);				\
    /* TODO: report the value of V, to extend the table in the future */ \
  } else {								\
    /* otherwise get them from the table */				\
    values=tab+iV*ntab;							\
  } /* if iV .. else */

/************************/
/* tabulated GATES by Rush-Larsen */
#define DOTGATES \
  a=values;								\
  b=values+nt;								\
  for (it=0;it<nt;it++) {						\
    gateold=tgate[it];							\
    tgate[it]=a[it]+b[it]*gateold; /* this is the RL step */		\
    if (!isfinite(tgate[it])) {						\
      URGENT_MESSAGE("\nNAN (not-a-number) at t=%ld x=%d y=%d z=%d tab gate %d\n",t,x,y,z,it); \
      URGENT_MESSAGE("This happened while multiplying %g by %g and adding %g\n",gateold,b[it],a[it]); \
      ABORT("");							\
    } /* if !isfinite */						\
  } /* for it */

/*************************************************************/
/* OTHER VARIABLES by forward Euler */
#define DOOTHER \
  if (!fddt(u,nv,values,ntab,p,var,du,no,nalp,nbet,nn)) {		\
    URGENT_MESSAGE("error calculating fddt(%s) at t=%ld point %d %d %d: u=",ionic,t,x,y,z); \
    for(iv=0;iv<nv;iv++) URGENT_MESSAGE(" %lg",u[iv]);			\
    ABORT("\n");							\
  } /*  if !fddt... */							\
  for (io=0;io<no;io++) {						\
    u[io] += ht*du[io]; /* this is the FE step */			\
    if (u[io]!=u[io]) {							\
      URGENT_MESSAGE("\nNAN (not-a-number) detected at t=%ld x=%d y=%d z=%d v=%d\n",t,x,y,z,io); \
      URGENT_MESSAGE("This happened while incrementing ");		\
      for(iv=0;iv<nv;iv++) URGENT_MESSAGE("%c%lg",iv?',':'(',u[iv]);	\
      URGENT_MESSAGE(") by %lg*",ht);					\
      for(jo=0;jo<no;jo++) URGENT_MESSAGE("%c%lg",jo?',':'(',du[jo]);	\
      URGENT_MESSAGE(")\n");						\
      ABORT("");							\
    } /* if NaN */							\
  } /* for io */

/*************************************************************/
/* NONTAB GATES by Rush-Larsen                               */
#define DONGATESRL \
  for (in=0;in<nn;in++) {						\
    a=nalp+in;								\
    b=nbet+in;								\
    calcab(a,b,ht);							\
    gateold=ngate[in];							\
    ngate[in]=(*a)+(*b)*gateold; /* this is the RL step */		\
    if (!isfinite(ngate[in])) {						\
      URGENT_MESSAGE("\nNAN (not-a-number) at t=%ld x=%d y=%d z=%d nontab gate %d\n",t,x,y,z,in); \
      URGENT_MESSAGE("This happened while multiplying %g by %g and adding %g\n",gateold,*b,*a); \
      ABORT("");							\
    } /* if !isfinite */						\
  } /* for in */
  
/*************************************************************/
/* NONTAB GATES by forward Euler                             */
#define DONGATESFE \
  for (in=0;in<nn;in++) {						\
    a=nalp+in;								\
    b=nbet+in;								\
    gateold=ngate[in];							\
    ngate[in] = gateold+ht*((*a)-((*a)+(*b))*gateold); /* this is the FE step */ \
    if (!isfinite(ngate[in])) {						\
      URGENT_MESSAGE("\nNAN (not-a-number) at t=%ld x=%d y=%d z=%d nontab gate %d\n",t,x,y,z,in); \
      URGENT_MESSAGE("This happened at gateold=%g ht=%g alpha=%g beta=%g\n",gateold,ht,*b,*a); \
      ABORT("");							\
    } /* if !isfinite */						\
  } /* for in */

enum {
  tgo,
  tog,
  totg,
  numorders
} ordertype;

/****************/
RUN_HEAD(rushlarsen) {
  IONIC_CONST(Par,p);			/* set of ionic cell parameters */
  IONIC_CONST(Var,var);		/* description of variable parameters */
  IONIC_CONST(IonicFtab *,ftab);	/* the tabulated functions calculator */
  IONIC_CONST(IonicFddt *,fddt);	/* the rhs functions except gates */
  IONIC_CONST(int,no);			/* number of non-gate variables in the state vector */
  IONIC_CONST(int,nn);			/* num of nontabulated gate variables in the state vector */
  IONIC_CONST(int,nt);			/* num of tabulated gate variables in the state vector */
  IONIC_CONST(int,ntab);		/* number of tabulated functions */
  IONIC_CONST(int,V_index);		/* index of V in state vector */
  DEVICE_ARRAY(char,ionic);		/* name of the ionic cell model */
  DEVICE_CONST(real,ht);		/* the time step */
  DEVICE_ARRAY(char,order);		/* text code of the order of execution */
  DEVICE_CONST(int,whichorder);		/* numeric code of the order of execution */
  DEVICE_CONST(int,exp_ngate);		/* if nonzero, non-tabulated gates are steppedn by RL, otherwise FE */
  DEVICE_CONST(int,nV);			/* number of 'rows' in the table */
  DEVICE_CONST(real,Vmin);		/* minimal value of V in the table */
  DEVICE_CONST(real,one_o_dV);		/* the inverse of increment of V in the table */
  DEVICE_ARRAY(real,du/*[no]*/);	/* array of right-hand sides for the FE part */
  DEVICE_ARRAY(real,nalp/*[nn]*/);	/* array of alphas for the nontab RL part */
  DEVICE_ARRAY(real,nbet/*[nn]*/);	/* array of betas for the nontab RL part */
  DEVICE_ARRAY(real,tab/*[ntab*nV]*/);	/* the table of function values */
  DEVICE_ARRAY(real,adhoc/*[ntab]*/);	/* freshly calculated values of functions */
  int nv;				/* size of the state vector */
  int x, y, z;				/* space grid counters */
  int io, jo, in, it, it1, iv;		/* vector components counters */
  int iV;				/* table row counter */
  real V;				/* transmembrane voltage value */
  real *u;				/* state vector at this point */
  real *ngate;				/* subvector of nontab gate values */
  real *tgate;				/* subvector of tab gate values */
  real *values;				/* vector of values of tabulated functions */
  real *a, *b;				/* pointers to subvectors of Rush-Larsen step coefficients */
  real gateold;				/* aux variable */
  
  /* The number of layers given to this device */
  nv=s.v1-s.v0+1;
  /* This should coincide with that required by the ionic model */
  ASSERT(no+nn+nt==nv);
  
  for(x=s.x0;x<=s.x1;x++) {
    for(y=s.y0;y<=s.y1;y++) {
      for(z=s.z0;z<=s.z1;z++) {
	if(isTissue(x,y,z)){
	  u = (real *)(New + ind(x,y,z,s.v0));
	  ngate=u+no;
	  tgate=u+no+nn;
	  switch (whichorder) {
	  case tgo:
	    DOTABLES;
            DOTGATES;
	    DOOTHER;
	    DONGATESFE;
	    break;
	  case tog:
	    DOTABLES;
	    DOOTHER;
	    DONGATESFE;
            DOTGATES;
	    break;
	  case totg:
	    DOTABLES;
	    DOOTHER;
	    DONGATESFE;
	    DOTABLES;
            DOTGATES;
	    break;
	  case tgo+numorders:
	    DOTABLES;
            DOTGATES;
	    DOOTHER;
	    DONGATESRL;
	    break;
	  case tog+numorders:
	    DOTABLES;
	    DOOTHER;
	    DONGATESRL;
            DOTGATES;
	    break;
	  case totg+numorders:
	    DOTABLES;
	    DOOTHER;
	    DONGATESRL;
	    DOTABLES;
            DOTGATES;
	    break;
	  default:
	    EXPECTED_ERROR("unknown order of execution code %d\n",whichorder);
	  } /* switch whichorder */
	} /*  if isTissue */
      } /*  for z */
    } /*  for y */
  } /*  for x */
} RUN_TAIL(rushlarsen)

  /**********************/
DESTROY_HEAD(rushlarsen) {
  FREE(S->du);
  FREE(S->nalp);
  FREE(S->nbet);
  FREE(S->adhoc);
  FREE(S->tab);
  FREE(S->u);
  if (S->I.var.n) {
    FREE(S->I.var.src);
    FREE(S->I.var.dst);
  }
  FREE(S->I.p);
} DESTROY_TAIL(rushlarsen)

/* Declare all available ionic models */
#define D(a) IonicFtab ftab_##a;
#include "ioniclist.h"
#undef D
#define D(a) IonicFddt fddt_##a;
#include "ioniclist.h"
#undef D
#define D(a) IonicCreate create_##a;
#include "ioniclist.h"
#undef D

/****************************************/
CREATE_HEAD(rushlarsen) {
  int nv=dev->s.v1-dev->s.v0+1;
  int no, nn, nt, ntab, nV;
  int step, iv, ix, iy, iz;
  int in, it, io, jo, iV;
  real *ufull, *u, *values;
  real V, alp, bet;

  /* Accept the time step */
  ACCEPTR(ht,RNONE,0.,RNONE);
  if (ht==0) MESSAGE("/* WARNING: ht=0 is formally allowed but hardly makes sense */");

  /* Accept the execution order */
  ACCEPTS(order,"totg");
  STRSWITCH(order);
  STRCASE("tgo")  S->whichorder=tgo;
  STRCASE("tog")  S->whichorder=tog;
  STRCASE("totg") S->whichorder=totg;
  STRDEFAULT EXPECTED_ERROR("\nrushlarsen: unknown execution order '%s'\n",order); 
  STRENDSW

  ACCEPTI(exp_ngate,0,0,1);
  if (exp_ngate) S->whichorder+=numorders;

  /* Accept the ionic cell model */
  ACCEPTS(ionic,NULL);
  {
    char *pars;
    MALLOC(pars,(long)MAXSTRLEN);
    BEGINBLOCK("par=",pars);
    S->u=NULL;
    #define D(a)							\
    if (0==stricmp(S->ionic,#a)) {					\
      if NOT(create_##a(&(S->I),pars,&(S->u),dev->s.v0))		\
    	EXPECTED_ERROR("reading parameters for %s in \"%s\"",S->ionic,pars); \
      no=S->I.no; nn=S->I.nn; nt=S->I.nt;				\
      ntab=S->I.ntab;							\
      ASSERT(no+nn+nt==nv);						\
      S->I.ftab=ftab_##a;						\
      S->I.fddt=fddt_##a;						\
    } else
    #include "ioniclist.h"
    #undef D

    EXPECTED_ERROR("unknown ionic model %s",S->ionic);
    ENDBLOCK;
    FREE(pars);
  }

  /* Allocated working arrays */
  CALLOC(S->du,no,sizeof(real));
  CALLOC(S->nalp,nn,sizeof(real));
  CALLOC(S->nbet,nn,sizeof(real));
  CALLOC(S->adhoc,ntab,sizeof(real));

  /* If the ionic model allocated initial state, possibly */
  /* finalize it by calculating stationary values of the gates */
  if (S->u) {
    ACCEPTI(equilibrate_gates,0,0,1);
    V=S->u[S->I.V_index];
    values=S->adhoc;

    if (!(S->I.fddt(S->u,nv,values,ntab,S->I.p,S->I.var,S->du,no,S->nalp,S->nbet,nn))) {
      URGENT_MESSAGE("error calculating fddt(%s) at parse time: u=",ionic); 
      for(iv=0;iv<nv;iv++) URGENT_MESSAGE(" %lg",u[iv]);			
      ABORT("\n");							
    } /*  if !fddt... */							
    for (in=0;in<nn;in++) {
      alp=S->nalp[in];
      bet=S->nbet[in];
      S->u[no+in]=alp/(alp+bet);
    } /* for in */
    if (!(S->I.ftab)(V,values,ntab))
      EXPECTED_ERROR("error calculating ftab(%s) for V=%g\n",S->ionic,V);
    for (it=0;it<nt;it++) {
      alp=values[it];
      bet=values[nt+it];
      S->u[no+nn+it]=alp/(alp+bet);
    } /* for it */
  } else {
    if (find_key("equilibrate_gates=",w)) {
      MESSAGE("\n/* Warning: equilibrate_gate parameter specified for a model which "
	      " does not define a standard state. "
	      "The parameter will be ignored. */\n");
    } /* if findkey */
  } /* if S->u else */
  
  /* In any case, do the tabulation */
  ACCEPTR(Vmin,-200,RNONE,RNONE);
  ACCEPTR(Vmax,+200,RNONE,RNONE);
  ACCEPTR(dV,0.01,RSUCC(0),RNONE);
  S->nV=nV=ceil((S->Vmax-S->Vmin)/S->dV);
  S->one_o_dV=1.0/S->dV;
  CALLOC(S->tab,ntab*nV,sizeof(real));

  for (iV=0;iV<nV;iV++) {
    V=Vmin+(iV+0.5)*dV;
    values=&(S->tab[ntab*iV]);
    if (!(S->I.ftab)(V,values,ntab))
      ABORT("error calculating ftab(%s) table for V=%g\n",S->ionic,V);
    for (it=0;it<nt;it++)
      calcab(values+it,values+nt+it,ht);
  }

  ACCEPTI(rest,0,0,INONE);
  if (S->rest) {
    if (S->u) MESSAGE("\n/* NOTICE: finding resting state while already defined by the rhs */");
    else CALLOC(S->u,nv,sizeof(real));

    /* Need full vector, in case variable parameters use extra layers.   */
    /* NB extra layers may be above as well as below this device's space */
    CALLOC(ufull,vmax,sizeof(real)); 
    u=&(ufull[dev->s.v0]);

    if (S->I.var.n) MESSAGE("\n/* NOTICE: finding resting state while parameters are variable */");

    for (step=0;step<S->rest;step++) {
      if (!(S->I.fddt(u,nv,values,ntab,S->I.p,S->I.var,S->du,no,S->nalp,S->nbet,nn))) {
	URGENT_MESSAGE("error calculating fddt(%s) at parse time, iteration %d: u=",ionic,step); 
	for(iv=0;iv<nv;iv++) URGENT_MESSAGE(" %lg",u[iv]);			
	ABORT("\n");							
      } /*  if !fddt... */							
      for (in=0;in<nn;in++) {
	alp=S->nalp[in];
	bet=S->nbet[in];
	u[no+in]=alp/(alp+bet); /* would it be better to do RL step? */
      } /* for in */
      if (!(S->I.ftab)(V,values,ntab))
	EXPECTED_ERROR("error calculating ftab(%s) for V=%g\n",S->ionic,V);
      for (it=0;it<nt;it++) {
	alp=values[it];
	bet=values[nt+it];
	u[no+nn+it]=alp/(alp+bet); /* would it be better to do RL step? */
      } /* for it */
      for (io=0;io<no;io++) {
	u[io]+=S->ht*S->du[io];
	if (!isfinite(u[io])) {
	  URGENT_MESSAGE("\nNAN (not-a-number) at parse time, iteration %d component %d: v=%d",step,io);
	  URGENT_MESSAGE("This happened after an increment by %lg*",S->ht);
	  for(jo=0;jo<nv;jo++) URGENT_MESSAGE("%c%lg",jo?',':'(',S->du[jo]);
	  URGENT_MESSAGE(")\n");
	  return 0;
	}
      } /* for io */
    } /* for step */

    /* copy what is needed */
    for (iv=0;iv<nv;iv++) S->u[iv]=u[iv];

    /* don't need this one any more */
    FREE(ufull);
  } /* if S->rest */

  if (S->u) {
    MESSAGE0("\n/* Resting state: "); for(iv=0;iv<nv;iv++) MESSAGE1("%lg ",(S->u)[iv]); MESSAGE0("*/");
    /* Fill up the whole of the grid with the resting state */
    #if MPI
    if (dev->s.runHere) {
    #endif
      for (ix=dev->s.x0;ix<=dev->s.x1;ix++) { 
	for (iy=dev->s.y0;iy<=dev->s.y1;iy++) {
	  for (iz=dev->s.z0;iz<=dev->s.z1;iz++) {
	    for (iv=dev->s.v0;iv<=dev->s.v1;iv++) {
	      New[ind(ix,iy,iz,iv)]=(S->u)[iv-dev->s.v0];
	    } /* for iv */
	  } /* for iz */
	} /* for iy */
      } /* for ix */
    #if MPI
    } /* if runHere */
    #endif
  } else {
    /* No resting state was defined */
    MESSAGE("\n/* Notice: no standard state defined for ionic model '%s' */\n",ionic);
    CALLOC(S->u,nv,sizeof(real));
  }
  
} CREATE_TAIL(rushlarsen,1)
